-- Some test examples for contracts.

{-# LANGUAGE TemplateHaskell #-}
{-# OPTIONS_GHC -XDatatypeContexts #-}
-- ghc needs option to allow H98 class context in a data type

module Development.ContractTest where

import Development.Contract


t1 = $assert (prop (<0) =: true) [3::Int]

t2 = $assert (true =: true |> pNil) []

nat :: Contract Integer
nat = prop (>=0)

nats :: Contract [Integer]
nats = list nat

t3 = $assert nats [10,9..(-2)]

atLeast :: Int -> Contract [a]
atLeast 0 = true
atLeast n = true =: (atLeast (n-1))

t4 = $assert (atLeast 4) [1,2,3]

t5 = $assert (pJust nat) (Just (-2))

absE = $attach abs (true >-> nat)


t6 = $attach Nothing $(p 'Nothing)

t7 = $attach (Just True) $(p 'Nothing)

t8 = $attach [3] $(p '[])

t9 = $attach [] ($(p '(:)) true $(p '[]))

t10 = $attach Nothing ($(p 'Just) true)

absC = $attach (+1) (true >-> nat)  -- faulty implementation

t11 = absC (-1)

t12 = absC (-2)

constC = $attach (\x y -> y `seq` x) (true >-> false >-> true) -- faulty impl

t13 = constC 4 7

sqrtC :: Float -> Float
sqrtC = $attach sqrt (prop (>=0) >-> prop (>= 0))

t14 = sqrtC 4

t15 = sqrtC (-4)

listOfLength :: Int -> Contract [a]
listOfLength 0 = pNil
listOfLength n = pCons true (listOfLength (n-1))

lengthAtLeast :: Int -> Contract [a]
lengthAtLeast 0 = true
lengthAtLeast n = pCons true (lengthAtLeast (n-1))

takeC :: Int -> [a] -> [a]
takeC = $attach take (prop (>= 0) >>-> \n -> lengthAtLeast n >-> listOfLength n)

t16 = takeC (-5) [1,2,3,4]

t17 = takeC 5 [1,2,3,4]

applyC = $attach ($) ((nat >-> nat) >>-> \f -> prop (< f (-2)) >-> nat)

t18 = applyC (+1) 42 

t19 = $assert ($(p 'Left) true) (Left 'a')

t20 = $assert ($(p 'Left) true) (Right 'a')

-- tree data type

data Ord a => Tree a = Leaf a | Tree a :^: Tree a
  deriving Show 

-- pLeaf :: Contract a -> Contract (Tree a)
-- pLeaf = $(p 'Leaf)

-- pNode :: Ord a => Contract (Tree a) -> Contract (Tree a) -> Contract (Tree a)
-- pNode = $(p '(:^:))

$(deriveContracts ''Tree)

tree :: Ord a => Contract a -> Contract (Tree a)
tree c = pLeaf c |> (tree c =:^: tree c)

o1 = $assert (pLeaf true =:^: pLeaf true) ((Leaf 4 :^: Leaf 5) :^: Leaf 7)

o2 = $assert pNotLeaf (Leaf 7)

o3 = $assert (tree nat) ((Leaf 4 :^: Leaf 5) :^: Leaf (-7))


-- logical formulae

data Form = Imp Form Form | And Form Form |
            Or Form Form | Not Form | Atom Char
  deriving Show

$(deriveContracts ''Form)

conjNF, disj, lit, atom :: Contract Form

conjNF = pAnd conjNF conjNF |> disj
disj = pOr disj disj |> lit
lit = Development.ContractTest.pNot atom |> atom
atom = pAtom true

right = pImp (right & pNotImp) right |>
        pAnd (right & pNotAnd) right |>
        pOr (right & pNotOr) right |>
        Development.ContractTest.pNot right |> pAtom true

rightConjNF :: Contract Form
rightConjNF = conjNF & right

f1 = Not (Atom 'a')
f2 = Not f1
f3 = And (Or (Atom 'a') f1) (Or f1 (Atom 'b'))
f4 = Or f1 f3
f5 = Or (Atom 'a') (And (Atom 'b') (Atom 'c'))
f6 = And (Atom 'a') (And (Atom 'b') (Atom 'c')) 
f7 = And (Atom 'a') (Or (Atom 'b') f2)

clausalNF = $attach clausalNF' (rightConjNF >-> list (list lit))

clausalNF' :: Form -> [[Form]]
clausalNF' (And f1 f2) = clause f1 : clausalNF' f2
clausalNF' f = [clause f]

clause = $attach clause' ((disj & right) >-> list lit)

clause' :: Form -> [Form]
clause' (Or f1 f2) = f1 : clause' f2
clause' lit = [lit] 

-- Bad contract, because function does not check user input.
getNat = $attach getNat' (io nat)

getNat' :: IO Integer
getNat' = do
  xs <- getLine
  return (read xs)